/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Clippings extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

var gClippingsList, gStatusBar;
var gMoveToMnu1, gMoveToMnu2, gCopyToMnu1, gCopyToMnu2;
var gCurrentListItemIndex = -1;
var gStrBundle;
var gAppUtils;
var gDataSource;
var gClippingsListener;
var gIsFolderMenuSeparatorInitialized = {};

// Clippings XPCOM service
var gClippingsSvc;

// Undo
var gUndoStack = {
  length: 0,
  _stack: [],

  push: function (aState)
  {
    this._stack.push(aState);
    this.length++;
  },

  pop: function ()
  {
    var rv = this._stack.pop();
    this.length--;
    return rv;
  }
};

// Redo - only 1 undo action is reversible
var gRedoStack = {
  length:   0,
  _lastUndo: null,

  push: function (aState)
  {
    this._lastUndo = aState;
    this.length = (this.length == 0 ? 1 : 1);
  },

  pop: function ()
  {
    var rv = {};
    for (var ppty in this._lastUndo) {
      rv[ppty] = this._lastUndo[ppty];
    }
    this._lastUndo = null;
    this.length = 0;
    return rv;
  }
};


// Flags for gUndoStack._stack[i].action
const ACTION_EDITNAME = 1;
const ACTION_EDITTEXT = 2;
const ACTION_DELETECLIPPING = 3;
const ACTION_CREATENEW = 4;
const ACTION_CHANGEPOSITION = 5;
const ACTION_CREATENEWFOLDER = 6;
const ACTION_DELETEFOLDER = 7;
const ACTION_MOVETOFOLDER = 8;
const ACTION_COPYTOFOLDER = 9;
const ACTION_DELETEEMPTYFOLDER = 10;

// Flags for aDestUndoStack parameter of functions for reversible actions
const UNDO_STACK = 1;
const REDO_STACK = 2;



// Drag and drop tree list items to move them.
var dndStartPos = null;
var dndStartURI = null;
var dndExtText = null;

var listitemObserver = {
  onDragStart: function (event, transferData, action)
  {
    if (event.target.tagName != "treechildren") {
      return;
    }
    var index = gClippingsList.tree.boxObject.getRowAt(event.clientX,
						       event.clientY);
    var uri = gClippingsList.getURIAtIndex(index);
    var pos = gClippingsSvc.ctrIndexOf(uri);
    dndStartURI = uri;
    dndStartPos = pos;

    transferData.data = new TransferData();
    transferData.data.addDataForFlavour("text/unicode", pos);
  },

  onDragOver:  function (event, flavour, session) {},
  onDragExit:  function (event, session) {},

  onDrop: function (event, transferData, session) 
  {
    if (transferData.data != "") {
      dndExtText = transferData.data;
    }
  },

  getSupportedFlavours: function () 
  {
    var flavours = new FlavourSet();
    flavours.appendFlavour("text/unicode");
    return flavours;
  } 
};


var treeBuilderObserver = {
  // Mozilla 1.7 (Firefox 1.0) only
  canDropBeforeAfter: function (idx, orient)
  {
    return true;
  },

  // Mozilla 1.7 (Firefox 1.0) only
  canDropOn: function (idx, orient)
  {
    return true;
  },

  canDrop: function (idx, orient)
  {
    // Return true to allow dropping into a folder.  Return the `orient' param
    // to allow dropping only if the folder is expanded in the tree list.
    return true;
  },

  onDrop: function (idx, orient)
  {
    var uri = gClippingsList.getURIAtIndex(idx);
    var newPos = gClippingsSvc.ctrIndexOf(uri);
    // `orient' is 0 (zero) if dragging and dropping into a folder item.
    var destParent = orient == 0 ? uri : gClippingsSvc.getParent(uri);

    gAppUtils.log("DnD orientation: " + orient + "\nnewPos: " + newPos);

    if (! dndStartURI) {
      // Creating a new clipping from external text that is dragged and dropped
      // into Clippings Manager.
      if (dndExtText) {
	if (orient == 1) {  // nsIXULTreeBuilderObserver.DROP_AFTER
	  newPos++;
	}
	else if (orient == 0) {  // nsIXULTreeBuilderObserver.DROP_ON
	  newPos = null;  // Will append to end of folder being dropped to.
	}
	var name = gClippingsSvc.createClippingNameFromText(dndExtText, 0);
	var text = dndExtText;
	gClippingsSvc.createNewClippingEx(destParent, null, name, text, newPos, false);
      }
      else {
	gAppUtils.beep();
	gAppUtils.log("Unknown item dropped into tree list; ignoring.");
      }
      this._endDnD();
      return;
    }

    var srcParent = gClippingsSvc.getParent(dndStartURI);

    if (srcParent == destParent) {
      if (orient == 0) {  // nsIXULTreeBuilderObserver.DROP_ON
	// Dragging and dropping a folder item into its containing folder.
	gAppUtils.beep();
	gAppUtils.log("The selected item already belongs in this folder!");
	this._endDnD();
	return;
      }
      else if (orient == -1) { // nsIXULTreeBuilderObserver.DROP_BEFORE
	if (dndStartPos < newPos) newPos--;
      }
      else if (orient == 1) {  // nsIXULTreeBuilderObserver.DROP_AFTER
	if (dndStartPos > newPos) newPos++;
      }

      gAppUtils.log(String.format("Orientation: %d\nChanging position from %d to %d", orient, dndStartPos, newPos));

      if (dndStartPos != newPos) {
	moveEntry(dndStartURI, srcParent, dndStartPos, newPos, UNDO_STACK);
	gClippingsList.selectedURI = dndStartURI;
      }
    }
    else {
      if (dndStartURI == destParent) {
	gAppUtils.beep();
	gAppUtils.log("Cannot move a folder into itself!");
	this._endDnD();
	return;
      }

      // Prevent infinite recursion due to moving a folder into its own subfolder.
      if (gClippingsSvc.isFolder(dndStartURI)) {
	var parentURI = gClippingsSvc.getParent(destParent);
	while (parentURI && parentURI != gClippingsSvc.kRootFolderURI) {
	  if (parentURI == dndStartURI) {
	    gAppUtils.beep();
	    gAppUtils.log("Cannot move a folder into a subfolder of itself!");
	    this._endDnD();
	    return;
	  }
	  parentURI = gClippingsSvc.getParent(parentURI);
	}
      }

      if (orient == 1) {
	// nsIXULTreeBuilderObserver.DROP_AFTER
	newPos++;
      }
      else if (orient == 0) {
	// Dragging and dropping into a folder - append the item to be
	// moved as the last item of the folder.
	newPos = null;
      }

      gAppUtils.log(String.format("Orientation: %d\nMoving to folder %S at position %d", orient, gClippingsSvc.getName(destParent), newPos));

      moveToFolderHelper(dndStartURI, srcParent, destParent, null, newPos,
			 UNDO_STACK, true);
    }

    updateDisplay();
    this._endDnD();
  },

  _endDnD: function ()
  {
    dndStartPos = null;
    dndStartURI = null;
    dndExtText = null;
  },
  
  onSelectionChanged: function () {},
  onToggleOpenState: function (idx) {}
};


// Common dialogs

function doAlert(aMessage)
{
  var title = gStrBundle.getString('appName');
  gAppUtils.alertEx(title, aMessage);
}


function doConfirm(aMessage)
{
  var rv;
  var title = gStrBundle.getString('appName');
  rv = gAppUtils.confirmEx(title, aMessage);
  return rv;
}



function init() 
{
  gAppUtils = new AppUtils();

  try {
    gClippingsSvc = Components.classes["clippings@mozdev.org/clippings;1"].getService(Components.interfaces.nsIClippingsService);
  }
  catch (e) {
    doAlert(e);
  }

  var treeElt = document.getElementById("clippings-list");
  gClippingsList = new RDFTreeWrapper(treeElt);
  gClippingsList.tree.builder.addObserver(treeBuilderObserver);
  gStrBundle = document.getElementById("ae-clippings-strings");
  gStatusBar = document.getElementById("app-status");

  // HACK! Change the width of the outer <vbox>, not its inner <tree>. Invoking
  // window.getComputedStyle() on the tree element when unloading this window
  // will always return its width which is 4 pixels less than it actually is,
  // resulting in a gradually narrowing tree in successive sessions!
  var treeBox = document.getElementById("tree-box");
  treeBox.style.width = gAppUtils.getPref("clippings.clipmgr.treelist.width");

  // Clippings backup
  var backupDirURL = gAppUtils.getCurrentProfileURL() + gAppUtils.BACKUP_DIR_NAME;
  gClippingsSvc.setBackupDir(backupDirURL);
  gClippingsSvc.setMaxBackupFiles(gAppUtils.getPref('clippings.backup.maxfiles'));

  var dsURL = gAppUtils.getCurrentProfileURL() + gAppUtils.CLIPDAT_FILE_NAME;
  var recovered = false;
  try {
    var ds = gClippingsSvc.getDataSource(dsURL);
  }
  catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
    doAlert(gStrBundle.getString("errorOutOfMemory"));
    return;
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorAccessDenied"),
			  gAppUtils.CLIPDAT_FILE_NAME));
    return;
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorFileLocked"),
			  gAppUtils.CLIPDAT_FILE_NAME));
    return;
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorFileTooBig"),
			  gAppUtils.CLIPDAT_FILE_NAME));
    return;
  }
  catch (e) {
    ds = doRecovery(dsURL);
    if (! ds) {
      // Recovery failed
      return;
    }
    recovered = true;
  }

  gClippingsList.tree.database.AddDataSource(ds);
  gDataSource = ds;

  var numEntries = gClippingsSvc.numItems;
  var deck = document.getElementById("entry-properties");
  deck.selectedIndex = numEntries == 0 ? 1 : 2;

  gClippingsList.tree.builder.rebuild();

  if (numEntries > 0) {
    gClippingsList.selectedIndex = 0;
    gClippingsList.tree.click();
    gClippingsList.tree.focus();
    gCurrentListItemIndex = 0;
  }
  
  if (recovered) {
    gAppUtils.beep();
    gStatusBar.label = gStrBundle.getString('recoverySuccess');
  }
  else {
    gStatusBar.label = numEntries + " " + gStrBundle.getString('numEntriesPost');
  }

  gClippingsListener = {
    origin: gClippingsSvc.ORIGIN_CLIPPINGS_MGR,

    newClippingCreated: function (aClippingURI)
    {
      detectExternallyCreatedEntry(aClippingURI);
    },

    newFolderCreated: function (aFolderURI) {
      detectExternallyCreatedFolder(aFolderURI);
    },

    importDone: function (aNumItems) {}
  };
  gClippingsSvc.addListener(gClippingsListener);

  gMoveToMnu1 = document.getElementById("move-to-1");
  gCopyToMnu1 = document.getElementById("copy-to-1");
  gMoveToMnu2 = document.getElementById("move-to-2");
  gCopyToMnu2 = document.getElementById("copy-to-2");

  buildCopyToMenu(gCopyToMnu1);
  buildCopyToMenu(gMoveToMnu1);
  buildCopyToMenu(gCopyToMnu2);
  buildCopyToMenu(gMoveToMnu2);

  // Preloading menu icons
  var image1 = document.createElement("menuitem");
  var image2 = document.createElement("menuitem");
  image1.setAttribute("image", "chrome://clippings/skin/images/folder.png");
  image2.setAttribute("image", "chrome://clippings/skin/images/clippings-root.png");
}


function detectExternallyCreatedEntry(aNewClippingURI)
{
  gClippingsList.selectedURI = aNewClippingURI;

  // No effect if parent folder of new clipping isn't expanded.
  gClippingsList.ensureURIIsVisible(aNewClippingURI); 

  gClippingsList.tree.click();

  var state = {
    action: ACTION_CREATENEW, 
    uri:    aNewClippingURI, 
    name:   gClippingsSvc.getName(aNewClippingURI),
    text:   gClippingsSvc.getText(aNewClippingURI),
    parentFolderURI: gClippingsSvc.getParent(aNewClippingURI)
  };
  gUndoStack.push(state);
  gAppUtils.log(String.format("New entry named %S added to undo stack (created outside Clippings Manager!)", state.name));

  var numEntries = gClippingsSvc.numItems;
  gStatusBar.label = numEntries + " " + gStrBundle.getString("numEntriesPost");
}


function detectExternallyCreatedFolder(aNewFolderURI)
{
  gClippingsList.selectedURI = aNewFolderURI;
  gClippingsList.ensureURIIsVisible(aNewFolderURI);
  gClippingsList.tree.click();

  var state = {
    action: ACTION_CREATENEWFOLDER,
    uri:    aNewFolderURI,
    name:   gClippingsSvc.getName(aNewFolderURI),
    pos:    gClippingsSvc.indexOf(aNewFolderURI),
    parentFolderURI: gClippingsSvc.getParent(aNewFolderURI)
  };

  gUndoStack.push(state);
  gAppUtils.log(String.format("New folder named %S added to undo stack (created outside Clippings Manager!)", state.name));

  removeFolderMenuSeparator();
  gMoveToMnu1.builder.rebuild();
  gCopyToMnu1.builder.rebuild();
  gMoveToMnu2.builder.rebuild();
  gCopyToMnu2.builder.rebuild();

  var numEntries = gClippingsSvc.numItems;
  gStatusBar.label = numEntries + " " + gStrBundle.getString("numEntriesPost");
}


function buildCopyToMenu(aMenu)
{
  // This function is invoked during Clippings Manager initialization.
  // Do not invoke it again afterwards!
  if (! gDataSource) {
    gAppUtils.log("Failed to initialize Copy To or Move To menu - data source not initialized!");
    return;
  }

  aMenu.database.AddDataSource(gDataSource);
  aMenu.builder.rebuild();
}


// This is called in Clippings Manager window's unload event handler
function unload()
{
  // See HACK! note in function init() for why we're getting the width from the
  // tree's enclosing <vbox> element, and not the tree itself.
  var treeBox = document.getElementById("tree-box");
  var style = window.getComputedStyle(treeBox, null);
  gAppUtils.setPref("clippings.clipmgr.treelist.width", style.width);
  
  gClippingsList.tree.builder.removeObserver(treeBuilderObserver);
  treeBuilderObserver = null;

  gClippingsSvc.purgeDetachedItems();

  try {
    gClippingsSvc.flushDataSrc();
  }
  catch (e if e.result == Components.results.NS_ERROR_NOT_INITIALIZED) {
    doAlert(gStrBundle.getString("errorSaveFailedDSNotInitialized"));
  }
  catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
    doAlert(gStrBundle.getString("errorOutOfMemory"));
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorAccessDenied"),
			  gAppUtils.CLIPDAT_FILE_NAME));
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorFileLocked"),
			  gAppUtils.CLIPDAT_FILE_NAME));
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorFileTooBig"),
			  gAppUtils.CLIPDAT_FILE_NAME));
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_READ_ONLY) {
    doAlert(String.format("%s: %s", gStrBundle.getString('errorFileReadOnly'),
			  gAppUtils.CLIPDAT_FILE_NAME));
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_DISK_FULL) {
    doAlert(String.format("%s: %s", gStrBundle.getString("errorDiskFull"),
			  gAppUtils.CLIPDAT_FILE_NAME));
  }
  catch (e) {
    doAlert(gStrBundle.getString("alertSaveFailed"));
  }

  gClippingsSvc.removeListener(gClippingsListener);
  gClippingsListener = null;
}


function doRecovery(aDataSrcURL)
{
  doAlert(gStrBundle.getString('recoverFromCorruption'));

  try {
    var recoveredDataSrc = gClippingsSvc.recoverFromBackup();
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_NOT_FOUND) {
    var choice = doConfirm(gStrBundle.getString('msgNoBackupDoImportOption'));
    if (! choice) {
      doAlert(gStrBundle.getString('errorRecoveryIncomplete'));
      return false;
    }

    try {
      gClippingsSvc.killDataSrc();
    }
    catch (e) {
      doAlert(gStrBundle.getString('errorCannotDeleteDSFile'));
      return false;
    }

    try {
      recoveredDataSrc = gClippingsSvc.getDataSource(aDataSrcURL);
    }
    catch (e) {
      doAlert(gStrBundle.getString('errorCannotRegenerateDS'));
      return false;
    }

    if (! doImport()) {
      return false;
    }
  }
  catch (e) {
    doAlert(gStrBundle.getString('errorRecoveryFailed'));
    return false;
  }

  return recoveredDataSrc;
}


function initFolderMenuSeparator(aMenuPopup)
{
  var menuID = aMenuPopup.parentNode.id;
  if (gIsFolderMenuSeparatorInitialized[menuID]) {
    return;
  }

  if (gClippingsSvc.getCountSubfolders(gClippingsSvc.kRootFolderURI) > 0) {
    var clippingsRootMnuItemID = menuID + "-root";
    var clippingsRootMnuItem = document.getElementById(clippingsRootMnuItemID);
    var sep = document.createElement("menuseparator");
    sep.id = menuID + "-separator";
    aMenuPopup.insertBefore(sep, clippingsRootMnuItem.nextSibling);

    gIsFolderMenuSeparatorInitialized[menuID] = true;
  }
}


function removeFolderMenuSeparator() 
{
  // This function should be called immediately before rebuilding the Move To
  // and Copy To menus, usually after creating a new folder, deleting a
  // clipping or folder, etc.
  for (var menuID in gIsFolderMenuSeparatorInitialized) {
    var popup = document.getElementById(menuID + "-popup");
    var sep = document.getElementById(menuID + "-separator");
    if (sep) {
      popup.removeChild(sep);
    }
    gIsFolderMenuSeparatorInitialized[menuID] = false;
  }
}


function newFolder() 
{
  updateCurrentClippingData();

  var parentFolderURI, pos;
  var selectedURI = gClippingsList.selectedURI;

  if (! selectedURI) {
    parentFolderURI = gClippingsSvc.kRootFolderURI;
  }
  else {
    parentFolderURI = gClippingsSvc.getParent(selectedURI);
    //pos = gClippingsSvc.indexOf(selectedURI);
  }

  var newFolderName = gStrBundle.getString('newFolderName');
  newFolderHelper(parentFolderURI, newFolderName, null, pos, UNDO_STACK);
}


function newFolderHelper(aParentFolderURI, aFolderName, aURI, aPos, aDestUndoStack)
{
  var newNodeURI = gClippingsSvc.createNewFolderEx(aParentFolderURI, aURI, aFolderName, aPos, false, gClippingsSvc.ORIGIN_CLIPPINGS_MGR);

  gClippingsList.selectedURI = newNodeURI;
  gClippingsList.ensureURIIsVisible(newNodeURI);
  gClippingsList.tree.click();

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }

  var folderName = document.getElementById("clipping-name");
  folderName.select();
  folderName.focus();

  // Undo / redo
  var state = {
    action:  ACTION_CREATENEWFOLDER, 
    uri:     newNodeURI, 
    name:    aFolderName,
    pos:     aPos,
    parentFolderURI: aParentFolderURI
  };

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
    gAppUtils.log(String.format("New entry %S added to undo stack", aFolderName));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
    gAppUtils.log(String.format("New entry %S added to redo stack", aFolderName));
  }

  removeFolderMenuSeparator();
  gMoveToMnu1.builder.rebuild();
  gCopyToMnu1.builder.rebuild();
  gMoveToMnu2.builder.rebuild();
  gCopyToMnu2.builder.rebuild();
}


function newClipping() 
{
  updateCurrentClippingData();

  var parentFolderURI, pos;
  var selectedURI = gClippingsList.selectedURI;

  if (! selectedURI) {
    parentFolderURI = gClippingsSvc.kRootFolderURI;
  }
  else {
    parentFolderURI = gClippingsSvc.getParent(selectedURI);
    //pos = gClippingsSvc.indexOf(selectedURI);
  }

  newClippingHelper(parentFolderURI,
		    gStrBundle.getString('newClippingName'), "", null, pos,
		    UNDO_STACK);
}


function newClippingHelper(aParentFolder, aName, aText, aURI, aPos, aDestUndoStack)
{
  var newNodeURI = gClippingsSvc.createNewClippingEx(aParentFolder, aURI, aName, aText, aPos, true);

  if (! newNodeURI) {
    alert(gStrBundle.getString('errorCannotCreate'));
    return;
  }
  
  gClippingsList.selectedURI = newNodeURI;
  gClippingsList.ensureURIIsVisible(newNodeURI);
  gClippingsList.tree.click();

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }

  var clippingName = document.getElementById("clipping-name");
  clippingName.select();
  clippingName.focus();

  var state = {action: ACTION_CREATENEW, uri: newNodeURI, name: aName, text: aText, parentFolderURI: aParentFolder, pos: aPos};
  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
    gAppUtils.log(String.format("New entry %S added to undo stack", aName));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
    gAppUtils.log(String.format("New entry %S added to redo stack", aName));
  }
}


function pasteClippingAsNew()
{
  updateCurrentClippingData();

  var txt = gAppUtils.getTextFromClipboard();
  if (! txt) {
    gAppUtils.beep();
    gStatusBar.label = gStrBundle.getString('errorNoClipboardContent');
    return;
  }

  var parentFolderURI;
  var selectedURI = gClippingsList.selectedURI;
  if (! selectedURI) {
    parentFolderURI = gClippingsSvc.kRootFolderURI;
  }
  else {
    parentFolderURI = gClippingsSvc.getParent(selectedURI);
  }

  var newNodeURI = gClippingsSvc.newClippingFromText(txt, false, window,
						     parentFolderURI, true);
  if (! newNodeURI) {
    alert(gStrBundle.getString('errorCannotCreate'));
    return;
  }

  gClippingsList.selectedURI = newNodeURI;
  gClippingsList.ensureURIIsVisible(newNodeURI);
  gClippingsList.tree.click();

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }

  var clippingName = document.getElementById("clipping-name");
  clippingName.select();
  clippingName.focus();

  gUndoStack.push({action: ACTION_CREATENEW, uri: newNodeURI,
	           parentFolderURI: parentFolderURI,
		   name: gClippingsSvc.getName(newNodeURI),
		   text: gClippingsSvc.getText(newNodeURI)});
  gAppUtils.log(String.format("New entry added to undo stack\nName = %S", 
			       gClippingsSvc.getName(newNodeURI)));
}


function copyToFolder(aDestFolderURI)
{
  // We have to do this, otherwise if the Move button was clicked, it will
  // have the sunken look again when the mouse is hovered over it later.
  window.setTimeout("copyToFolderEx('" + aDestFolderURI + "');", 10);
}


function copyToFolderEx(aDestFolderURI)
{
  updateCurrentClippingData();

  var destFolderURI = aDestFolderURI;
  var itemURI = gClippingsList.selectedURI;
  var parentFolderURI = gClippingsSvc.getParent(itemURI);

  // Prevent infinite recursion due to copying a folder into its own subfolder.
  if (gClippingsSvc.isFolder(itemURI)) {
    var parentURI = gClippingsSvc.getParent(aDestFolderURI);
    while (parentURI && parentURI != gClippingsSvc.kRootFolderURI) {
      if (parentURI == itemURI) {
	gAppUtils.beep();
	gAppUtils.log("Cannot copy a folder into a subfolder of itself!");
	return;
      }
      parentURI = gClippingsSvc.getParent(parentURI);
    }
  }

  copyToFolderHelper(itemURI, parentFolderURI, destFolderURI, null, null, UNDO_STACK, false);
}


function copyToFolderHelper(aItemURI, aSrcFolderURI, aDestFolderURI, aDestItemURI, aDestPos, aDestUndoStack, aSelectCopiedItem)
{
  var prevIndex = gClippingsList.selectedIndex;
  var newURI = gClippingsSvc.copyTo(aItemURI, aDestFolderURI, aDestItemURI,
				    aDestPos, false,
				    gClippingsSvc.ORIGIN_CLIPPINGS_MGR);

  gAppUtils.log(String.format("Copy completed.  Name of copied item: %S\nParent folder of copied item: %S", gClippingsSvc.getName(aItemURI), gClippingsSvc.getName(aDestFolderURI)));

  removeFolderMenuSeparator();
  gMoveToMnu1.builder.rebuild();
  gCopyToMnu1.builder.rebuild();
  gMoveToMnu2.builder.rebuild();
  gCopyToMnu2.builder.rebuild();

  // The Clippings tree context menu doesn't close automatically after the
  // copy or move operation.
  document.getElementById("clippings-list-context").hidePopup();
  document.getElementById("toolbar-move-popup").hidePopup();

  var state = {
    action: ACTION_COPYTOFOLDER,
    copyURI:    newURI,
    originalURI:    aItemURI,
    srcFolder: aSrcFolderURI
  };

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
  }

  if (aSelectCopiedItem) {
    gClippingsList.selectedURI = newURI;
    gClippingsList.ensureURIIsVisible(newURI);
  }
  else {
    var numItems = gClippingsList.getRowCount();
    if (prevIndex == numItems) {  // Moved item was last list item.
      gClippingsList.selectedIndex = numItems - 1;
      gClippingsList.ensureIndexIsVisible(numItems - 1);
    }
    else {
      gClippingsList.selectedIndex = prevIndex;
      gClippingsList.ensureIndexIsVisible(prevIndex);
    }
  }

  updateDisplay();
}


function moveToFolder(aDestFolderURI)
{
  // We have to do this, otherwise if the Move button was clicked, it will
  // have the pressed-down look again when the mouse is hovered over it later.
  window.setTimeout("moveToFolderEx('" + aDestFolderURI + "');", 10);
}


function moveToFolderEx(aDestFolderURI)
{
  updateCurrentClippingData();

  var destFolderURI = aDestFolderURI;
  var itemURI = gClippingsList.selectedURI;

  if (gClippingsSvc.isFolder(itemURI) && itemURI == destFolderURI) {
    gAppUtils.beep();
    gAppUtils.log("Cannot move a folder into itself!");
    return;
  }

  // It is pointless to move an item into the same folder.
  var parentFolderURI = gClippingsSvc.getParent(itemURI);
  if (parentFolderURI == destFolderURI) {
    gAppUtils.beep();
    gAppUtils.log("The source and destination folders are the same!");
    return;
  }

  // Prevent infinite recursion due to moving a folder into its own subfolder.
  if (gClippingsSvc.isFolder(itemURI)) {
    var parentURI = gClippingsSvc.getParent(aDestFolderURI);
    while (parentURI && parentURI != gClippingsSvc.kRootFolderURI) {
      if (parentURI == itemURI) {
	gAppUtils.beep();
	gAppUtils.log("Cannot move a folder into a subfolder of itself!");
	return;
      }
      parentURI = gClippingsSvc.getParent(parentURI);
    }
  }

  moveToFolderHelper(itemURI, parentFolderURI, destFolderURI, null, null, UNDO_STACK, false);
}


function moveToFolderHelper(aItemURI, aSrcFolderURI, aDestFolderURI, aDestItemURI, aDestPos, aDestUndoStack, aSelectMovedItem)
{
  var pos = gClippingsSvc.indexOf(aItemURI);
  var prevIndex = gClippingsList.selectedIndex;
  var newURI = gClippingsSvc.copyTo(aItemURI, aDestFolderURI, aDestItemURI,
				    aDestPos, true,
				    gClippingsSvc.ORIGIN_CLIPPINGS_MGR);

  gAppUtils.log(String.format("Move completed.  Name of moved item: %S\nParent folder the item was moved to: %S", gClippingsSvc.getName(newURI), gClippingsSvc.getName(aDestFolderURI)));

  removeFolderMenuSeparator();
  gMoveToMnu1.builder.rebuild();
  gCopyToMnu1.builder.rebuild();
  gMoveToMnu2.builder.rebuild();
  gCopyToMnu2.builder.rebuild();

  // The Clippings tree context menu doesn't close automatically after the
  // copy or move operation.
  document.getElementById("clippings-list-context").hidePopup();
  document.getElementById("toolbar-move-popup").hidePopup();

  var state = {
    action: ACTION_MOVETOFOLDER,
    uri:    newURI,
    originalFolder: aSrcFolderURI,
    originalURI: aItemURI,
    originalPos: pos
  };

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
  }

  if (aSelectMovedItem) {
    // The following will not work if the parent of the item to select is
    // not expanded.
    gClippingsList.selectedURI = newURI;
    gClippingsList.ensureURIIsVisible(newURI);
  }
  else {
    var numItems = gClippingsList.getRowCount();
    if (prevIndex == numItems) {  // Moved item was last list item.
      gClippingsList.selectedIndex = numItems - 1;
      gClippingsList.ensureIndexIsVisible(numItems - 1);
    }
    else {
      gClippingsList.selectedIndex = prevIndex;
      gClippingsList.ensureIndexIsVisible(prevIndex);
    }
  }

  updateDisplay();
}



function updateEditStatus() 
{
  gStatusBar.label = gStrBundle.getString("editEntry");
}


function saveClippings() 
{
  updateCurrentClippingData();
  gStatusBar.label = gStrBundle.getString("saveProgress");

  var msg = gStrBundle.getString("saveCompleted");;
  try {
    gClippingsSvc.flushDataSrc();
  }
  catch (e) {
    msg = gStrBundle.getString("errorSaveFailed");
  }
  gStatusBar.label = msg;
}


function deleteClipping() 
{
  if (gClippingsList.getRowCount() == 0) {
    gAppUtils.beep();
    gStatusBar.label = gStrBundle.getString("msgNothingToDelete");
    return;
  }

  updateCurrentClippingData();

  var uri = gClippingsList.selectedURI;
  if (! uri) {
    return;
  }

  deleteClippingHelper(uri, UNDO_STACK);
}


function deleteClippingHelper(aURI, aDestUndoStack)
  // NOTE - This deletes both clippings and folders
{
  var deletedItemIndex = gClippingsList.selectedIndex;
  var state;
  var pos = gClippingsSvc.indexOf(aURI);

  if (gClippingsSvc.isClipping(aURI)) {
    state = {
      action: ACTION_DELETECLIPPING,
      uri:    aURI, 
      name:   gClippingsSvc.getName(aURI),
      text:   gClippingsSvc.getText(aURI),
      pos:    pos,
      parentFolderURI: gClippingsSvc.getParent(aURI)
    };

    try {
      gClippingsSvc.remove(aURI);
    }
    catch (e) {
      alert(gStrBundle.getString("errorCannotDelete"));
      return;
    }
  }
  else if (gClippingsSvc.isFolder(aURI)) {
    if (gClippingsSvc.getCount(aURI) > 0) {
      gAppUtils.log("Removing non-empty folder");
      state = {
	action: ACTION_DELETEFOLDER,
	name:   gClippingsSvc.getName(aURI),
	parentFolderURI: gClippingsSvc.getParent(aURI)
      };

      try {
	gClippingsSvc.detachFromFolder(aURI);
      }
      catch (e) {
	alert(gStrBundle.getString("errorCannotDelete"));
	return;
      }
    }
    else {
      gAppUtils.log("Removing empty folder");
      state = {
	action: ACTION_DELETEEMPTYFOLDER,
	name:   gClippingsSvc.getName(aURI),
	parentFolderURI: gClippingsSvc.getParent(aURI)
      };

      try {
	gClippingsSvc.remove(aURI);
      }
      catch (e) {
	alert(gStrBundle.getString("errorCannotDelete"));
	return;
      }
    }

    state.uri = aURI;
    state.pos = pos;
  }

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
    gAppUtils.log(String.format("Deleted entry %S (URI %s added to undo stack)\nPosition: %d", state.name, state.uri, state.pos));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
    gAppUtils.log(String.format("Deleted entry %S (URI %s added to redo stack)", state.name, uri));
  }

  removeFolderMenuSeparator();
  gClippingsList.tree.builder.rebuild();

  var numEntries = gClippingsSvc.numItems;
  gStatusBar.label = numEntries + " " + gStrBundle.getString("numEntriesPost");

  var numItems = gClippingsList.getRowCount();
  if (numItems == 0) {
    document.getElementById("entry-properties").selectedIndex = 1;
    gCurrentListItemIndex = -1;
  }
  else if (deletedItemIndex == numItems) {  // Deleted item was last list item
    gClippingsList.selectedIndex = numItems - 1;
    gClippingsList.ensureIndexIsVisible(numItems - 1);
    updateDisplay();
  }
  else {
    gClippingsList.selectedIndex = deletedItemIndex;
    gClippingsList.ensureIndexIsVisible(deletedItemIndex);
    updateDisplay();
  }
}


function updateCurrentEntryStatus()
{
  if (gClippingsList.getRowCount() == 0) {
    return;
  }

  var numEntries = gClippingsSvc.numItems;
  gStatusBar.label = numEntries + " " + gStrBundle.getString("numEntriesPost");

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }
}


function updateDisplay(aSuppressUpdateSelection)
{
  if (gClippingsList.getRowCount() == 0) {
    return;
  }

  var clippingName = document.getElementById("clipping-name");
  var clippingText = document.getElementById("clipping-text");
  var clippingTextLabel = document.getElementById("clipping-text-label");

  var uri = gClippingsList.selectedURI;
  if (! uri) {
    // Nothing selected - so just select whatever is at the current index.
    var numRows = gClippingsList.getRowCount();
    if (gCurrentListItemIndex > (numRows - 1)) {
      gCurrentListItemIndex = numRows - 1;
    }

    gClippingsList.selectedIndex = gCurrentListItemIndex;
    uri = gClippingsList.selectedURI;
  }

  if (gClippingsSvc.isFolder(uri)) {
    clippingTextLabel.disabled = true;
    clippingText.disabled = true;
  }
  else {
    clippingTextLabel.disabled = false;
    clippingText.disabled = false;
  }

  clippingName.value = gClippingsSvc.getName(uri);
  clippingText.value = gClippingsSvc.getText(uri);

  gCurrentListItemIndex = gClippingsList.selectedIndex;

  if (! aSuppressUpdateSelection) {
    gClippingsList.ensureIndexIsVisible(gClippingsList.selectedIndex);
  }
}


function updateName(aName)
{
  if (gCurrentListItemIndex == -1) {
    return;
  }

  var uri = gClippingsList.getURIAtIndex(gCurrentListItemIndex);
  updateNameHelper(uri, aName, UNDO_STACK);
}


function updateNameHelper(aURI, aName, aDestUndoStack)
{
  var oldName = gClippingsSvc.getName(aURI);
  if (oldName == aName) {
    return;
  }

  var state = {action: ACTION_EDITNAME, uri: aURI, name: oldName, text: null};
  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(state);
    gAppUtils.log(String.format("In function updateNameHelper(): Old Entry name: %S\nOld entry name added to undo stack", oldName));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
    gAppUtils.log(String.format("In function updateNameHelper(): Old Entry name: %S\nOld entry name added to redo stack", oldName));
  }

  gClippingsSvc.setName(aURI, aName);
}


function updateText(aText)
{
  if (gCurrentListItemIndex == -1) {
    return;
  }
  var uri = gClippingsList.getURIAtIndex(gCurrentListItemIndex);
  if (gClippingsSvc.isFolder(uri)) {
    // Folders don't have a `text' predicate, so just ignore
    return;
  }

  updateTextHelper(uri, aText, UNDO_STACK);
}


function updateTextHelper(aURI, aText, aDestUndoStack)
{
  var oldText = gClippingsSvc.getText(aURI);
  if (oldText == aText) {
    return;
  }

  var state = {action: ACTION_EDITTEXT, uri: aURI, name: null, text: oldText};
  if (aDestUndoStack == UNDO_STACK) {
    // DEBUGGING
    var name = gClippingsSvc.getName(aURI);
    // END DEBUGGING
    gUndoStack.push({action:ACTION_EDITTEXT, uri:aURI, name:name, text:oldText});
    // We shouldn't care what the old name is.
    // Non-debug invocation to push() should be as follows:
    /***
	gUndoStack.push(ACTION_EDITTEXT, aURI, null, oldText);
    ***/
    gAppUtils.log(String.format("In function updateTextHelper(): Entry name: %S\nOld text before edit added to undo stack\noldText = %S", name, oldText));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(state);
    gAppUtils.log(String.format("In function updateTextHelper(): Added to redo stack: oldText == %S", oldText));
  }

  gClippingsSvc.setText(aURI, aText);
}


function updateCurrentClippingData()
{
  if (gClippingsSvc.numItems > 0) {
    updateName(document.getElementById("clipping-name").value);

    var uri = gClippingsList.getURIAtIndex(gCurrentListItemIndex);
    if (gClippingsSvc.isClipping(uri)) {
      updateText(document.getElementById("clipping-text").value);
    }
  }
}


function moveUp()
{
  var currentURI = gClippingsList.selectedURI;
  var p = gClippingsSvc.getParent(currentURI);
  var i = gClippingsSvc.indexOf(currentURI);
  var max = gClippingsSvc.getCount(p);

  if (1 < i && i <= max) {
    moveEntry(currentURI, p, i, i - 1, UNDO_STACK);
    gClippingsList.selectedURI = currentURI;
    updateDisplay();
  }
  else if (i == 1) {
    gAppUtils.beep();
    gStatusBar.label = "Already at first item of folder";
  }
}


function moveDown()
{
  var currentURI = gClippingsList.selectedURI;
  var p = gClippingsSvc.getParent(currentURI);
  var i = gClippingsSvc.indexOf(currentURI);
  var max = gClippingsSvc.getCount(p);

  if (1 <= i && i < max) {
    moveEntry(currentURI, p, i, i + 1, UNDO_STACK);
    gClippingsList.selectedURI = currentURI;
    updateDisplay();
  }
  else if (i == max) {
    gAppUtils.beep();
    gStatusBar.label = "Already at last item of folder";
  }
}


function moveEntry(aURI, aParentFolderURI, aOldPos, aNewPos, aDestUndoStack)
  // aOldPos and aNewPos are 1-based (NOT zero-based) indices
{
  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push({action: ACTION_CHANGEPOSITION, uri: aURI,
		     currPos: aNewPos, prevPos: aOldPos,
		     parentFolderURI: aParentFolderURI});
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push({action: ACTION_CHANGEPOSITION, uri: aURI,
		     currPos: aNewPos, prevPos: aOldPos,
		     parentFolderURI: aParentFolderURI});
  }

  gAppUtils.log(String.format("In function moveEntry(): aOldPos=%d; aNewPos=%d", aOldPos, aNewPos));

  gClippingsSvc.changePosition(aParentFolderURI, aOldPos, aNewPos);
}


function doExport()
{
  window.openDialog("chrome://clippings/content/export.xul", "export_dlg", "dialog,modal,centerscreen");
}


function doImport()
{
  var fp = Components.classes["@mozilla.org/filepicker;1"]
                     .createInstance(Components.interfaces.nsIFilePicker);
  fp.init(window, gStrBundle.getString("dlgTitleImportClippings"), fp.modeOpen);
  fp.appendFilter(gStrBundle.getString("rdfImportFilterDesc"), "*.rdf");

  var dlgResult = fp.show();
  if (dlgResult != fp.returnOK) {
    return false;
  }

  var url = fp.fileURL.QueryInterface(Components.interfaces.nsIURI).spec;
  var path = fp.file.QueryInterface(Components.interfaces.nsIFile).path;
  var dsURL = gAppUtils.getCurrentProfileURL() + gAppUtils.CLIPDAT_FILE_NAME;

  // Prevent attempt at importing data source file.
  if (url == dsURL) {
    doAlert(String.format("%s %S", gStrBundle.getString("errorCannotImportDSFile"), url));
    return false;
  }

  gStatusBar.label = gStrBundle.getString("importBegin");

  try {
    var numImported = gClippingsSvc.importFromFile(url, false);
  }
  catch (e if e.result == Components.results.NS_ERROR_NOT_INITIALIZED) {
    doAlert(gStrBundle.getString('alertImportFailedNoDS'));
    gStatusBar.label = "";
    return false;
  }
  catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
    doAlert(gStrBundle.getString("errorOutOfMemory"));
    gStatusBar.label = "";
    return false;
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
    doAlert(gStrBundle.getString("errorAccessDenied"));
    gStatusBar.label = "";
    return false;
  }
  catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
    doAlert(gStrBundle.getString("errorFileTooBig"));
    gStatusBar.label = "";
    return false;
  }
  catch (e) {
    gAppUtils.log(e);
    var err = String.format("%s %S\n%s",
			    gStrBundle.getString("alertImportFailed1"), path,
			    gStrBundle.getString("alertImportFailed2"));
    doAlert(err);
    gStatusBar.label = "";
    return false;
  }

  // Handle empty RDF files
  if (numImported == 0) {
    gStatusBar.label = gStrBundle.getString("msgNoItems");
    return true;
  }

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    if (gClippingsList.getRowCount() > 0) {
      deck.selectedIndex = 0;
      gClippingsList.selectedIndex = 0;
      gClippingsList.tree.click();
    }
  }    
  // Status bar msg is overwritten in listbox.click() call, so redisplay
  // status of import.
  gStatusBar.label = String.format("%s%s",
				   gStrBundle.getString("importBegin"),
				   gStrBundle.getString("importDone"));
  try {
    gClippingsSvc.flushDataSrc();
  }
  catch (e) {
    // Don't do anything for now - try again when closing Clippings Manager.
  }

  return true;
}


function initClippingsListPopup()
{
  var numListItems = gClippingsList.getRowCount();
  if (numListItems == 0) {
    return false;  // Don't show popup menu if list box is empty
  }

  return true;
}


function undo()
{
  updateCurrentClippingData();

  if (gUndoStack.length == 0) {
    gAppUtils.beep();
    gStatusBar.label = gStrBundle.getString("msgNothingToUndo");
    return;
  }

  var undo = gUndoStack.pop();
  var srcFolder;  // Used for undoing move or copy
  var pos;        // Used for undoing folder or clipping creation

  if (undo.action == ACTION_DELETECLIPPING) {
    gClippingsSvc.createNewClippingEx(undo.parentFolderURI, undo.uri,
				      undo.name, undo.text, undo.pos, true);
    gClippingsList.selectedURI = undo.uri;
    gClippingsList.ensureURIIsVisible(undo.uri);
    updateDisplay();

    gAppUtils.log(String.format("Deletion undo\nEntry name: %S; entry URI: %S", undo.name, undo.uri));

    gRedoStack.push({action: ACTION_DELETECLIPPING, uri: undo.uri,
		     name: undo.name, text: undo.text, pos: undo.pos});
  }
  else if (undo.action == ACTION_EDITTEXT) {
    updateTextHelper(undo.uri, undo.text, REDO_STACK);
    gClippingsList.selectedURI = undo.uri;
    updateDisplay();

    var clippingText = document.getElementById("clipping-text");
    clippingText.focus();
    clippingText.select();

    gAppUtils.log(String.format("Entry text edit undo\nEntry name: %S; entry URI: %S; oldText: %S", undo.name, undo.uri, undo.text));
  }
  else if (undo.action == ACTION_EDITNAME) {
    updateNameHelper(undo.uri, undo.name, REDO_STACK);
    gClippingsList.selectedURI = undo.uri;
    updateDisplay();

    var clippingName = document.getElementById("clipping-name");
    clippingName.focus();
    clippingName.select();

    gAppUtils.log(String.format("Entry name edit undo\nEntry name: %S; entry URI: %S", undo.name, undo.uri));
  }
  else if (undo.action == ACTION_CREATENEW) {
    gClippingsList.selectedURI = undo.uri;
    pos = undo.pos || gClippingsSvc.indexOf(undo.uri);
    deleteClippingHelper(undo.uri);
    gAppUtils.log(String.format("Entry creation undo\nEntry name: %S; entry URI: %S", undo.name, undo.uri));
    gRedoStack.push({action: ACTION_CREATENEW, uri: undo.uri, 
	             name: undo.name, text: undo.text, pos: pos,
		     parentFolderURI: undo.parentFolderURI});
  }
  else if (undo.action == ACTION_CREATENEWFOLDER) {
    gClippingsList.selectedURI = undo.uri;
    pos = undo.pos || gClippingsSvc.indexOf(undo.uri);
    deleteClippingHelper(undo.uri);
    gRedoStack.push({action: ACTION_CREATENEWFOLDER, name: undo.name,
	             pos: pos, uri: undo.uri,
		     parentFolderURI: undo.parentFolderURI});
    gAppUtils.log("Folder creation undone; folder name: \"" + undo.name + "\"; URI: \"" + undo.uri + "\"");
  }
  else if (undo.action == ACTION_DELETEFOLDER) {
    gClippingsSvc.reattachToFolder(undo.parentFolderURI, undo.uri, undo.pos);
    gClippingsList.selectedURI = undo.uri;
    gCurrentListItemIndex = gClippingsList.selectedIndex;
    updateDisplay();

    gRedoStack.push({action: ACTION_DELETEFOLDER, uri: undo.uri});
  }
  else if (undo.action == ACTION_DELETEEMPTYFOLDER) {
    newFolderHelper(undo.parentFolderURI, undo.name, undo.uri, undo.pos, null);
    gCurrentListItemIndex = gClippingsList.selectedIndex;
    updateDisplay();

    gRedoStack.push({action: ACTION_DELETEEMPTYFOLDER, uri: undo.uri});
  }
  else if (undo.action == ACTION_MOVETOFOLDER) {
    pos = undo.originalPos;
    try {
      srcFolder = gClippingsSvc.getParent(undo.uri);
      moveToFolderHelper(undo.uri, srcFolder, undo.originalFolder,
			 undo.originalURI, pos, REDO_STACK, true);
    } catch (e) {
      throw ("Exception occurred while attempting to undo item move:\n" + e);
    }
    gAppUtils.log("Move undone.");
  }
  else if (undo.action == ACTION_COPYTOFOLDER) {
    var destFolder = gClippingsSvc.getParent(undo.copyURI);
    deleteClippingHelper(undo.copyURI);
    gRedoStack.push({action:      ACTION_COPYTOFOLDER,
		     originalURI: undo.originalURI,
		     srcFolder:   undo.srcFolder,
		     destFolder:  destFolder,
		     copyURI: undo.copyURI}); 
    gAppUtils.log("Copy undone.");
  }
  else if (undo.action == ACTION_CHANGEPOSITION) {
    moveEntry(undo.uri, undo.parentFolderURI, undo.currPos, undo.prevPos, REDO_STACK);
    gClippingsList.selectedURI = undo.uri;
    updateDisplay();
  }
}


function reverseLastUndo()
{
  updateCurrentClippingData();

  if (gRedoStack.length == 0) {
    gAppUtils.beep();
    return;
  }

  var redo = gRedoStack.pop();
  var srcFolder;  // Used for redoing move or copy
  var pos;        // Used for redoing folder or clipping creation

  if (redo.action == ACTION_DELETECLIPPING) {
    gClippingsList.selectedURI = redo.uri;
    deleteClippingHelper(redo.uri, UNDO_STACK);
    gAppUtils.log(String.format("Entry deletion redone!\nEntry name: %S; entry URI: %S", redo.name, redo.uri));
  }
  else if (redo.action == ACTION_EDITNAME) {
    updateNameHelper(redo.uri, redo.name, UNDO_STACK);
    gClippingsList.selectedURI = redo.uri;
    updateDisplay();

    var clippingName = document.getElementById("clipping-name");
    clippingName.focus();
    clippingName.select();

    gAppUtils.log(String.format("Entry name edit redone!\nEntry name: %S; entry URI:", redo.name, redo.uri));
  }
  else if (redo.action == ACTION_EDITTEXT) {
    updateTextHelper(redo.uri, redo.text, UNDO_STACK);
    gClippingsList.selectedURI = redo.uri;
    updateDisplay();

    var clippingText = document.getElementById("clipping-text");
    clippingText.focus();
    clippingText.select();

    gAppUtils.log(String.format("Entry text edit redone!\nEntry name: %S; entry URI: %S; oldText: %S", redo.name, redo.uri, redo.text));
  }
  else if (redo.action == ACTION_CREATENEW) {
    pos = null;  //redo.pos;
    newClippingHelper(redo.parentFolderURI, redo.name, redo.text, redo.uri, pos, UNDO_STACK);
    gClippingsList.selectedURI = redo.uri;
    gAppUtils.log(String.format("Deletion redone!\nEntry name: %S; entry URI: %S", redo.name, redo.uri));
  }
  else if (redo.action == ACTION_CREATENEWFOLDER) {
    pos = null;  //redo.pos;
    newFolderHelper(redo.parentFolderURI, redo.name, redo.uri, pos, UNDO_STACK);
    gClippingsList.selectedURI = redo.uri;
    gAppUtils.log("Folder creation redone!  Folder name: \"" + redo.name + "\"");
  }
  else if (redo.action == ACTION_DELETEFOLDER) {
    deleteClippingHelper(redo.uri, UNDO_STACK);
  }
  else if (redo.action == ACTION_DELETEEMPTYFOLDER) {
    deleteClippingHelper(redo.uri, UNDO_STACK);
  }

  else if (redo.action == ACTION_MOVETOFOLDER) {
    try { 
      srcFolder = gClippingsSvc.getParent(redo.uri);
      moveToFolderHelper(redo.uri, srcFolder, redo.originalFolder,
			 redo.originalURI, null, UNDO_STACK, false);
    } catch (e) {
      throw ("Exception occurred while attempting to redo item move:\n" + e);
    }
    gAppUtils.log("Move redone!");
  }
  else if (redo.action == ACTION_COPYTOFOLDER) {
    copyToFolderHelper(redo.originalURI, redo.srcFolder, redo.destFolder,
		       redo.copyURI, null, UNDO_STACK, false);
    gAppUtils.log("Copy redone!");
  }
  else if (redo.action == ACTION_CHANGEPOSITION) {
    moveEntry(redo.uri, redo.parentFolderURI, redo.currPos, redo.prevPos, UNDO_STACK);
    gClippingsList.selectedURI = redo.uri;
    updateDisplay();
  }

  gStatusBar.label = gStrBundle.getString("redoStatus");
}


function showHelp() 
{
  updateCurrentClippingData();
  doAlert(gStrBundle.getString("helpMsg"));
}
