/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Clippings extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

//
// Integration with host application
//

var gClippingsExec = {
  dataSrcInitialized: false,
  showDialog:    true,
  appUtils:      null,
  clippingsSvc:  null,
  strBundle:     null,
  listener:      null,

  newFromClipboard: function () 
  {
    if (! this.dataSrcInitialized) {
      // The RDF data source has to be initialized if it has not already been
      // done so, otherwise RDF node creation will fail, and the new Clippings
      // entry will never be created.
      // This initialization is done in the code for the Clippings popup menu's
      // `onpopupshowing' event handler.
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    var txt = this.appUtils.getTextFromClipboard();
    if (! txt) {
      alert(this.strBundle.getString('errorNoClipboardContent'));
      return;
    }

    var result = this.clippingsSvc.newClippingFromText(txt, this.showDialog, window, null, false);
    if (result) {
      window.setTimeout("gClippingsExec.saveClippings();", 1);
    }
  },

  newFromTextbox: function ()
  {
    if (! this.dataSrcInitialized) {
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    var cxtMenu = document.getElementById("contentAreaContextMenu");
    var result;

    // ** FIREFOX 2 **
    // gContextMenu.onTextInput is also true inside a rich edit box!
    if (gContextMenu.onTextInput) {

      // Must close content area context menu manually - otherwise it will
      // reappear again after creating new folder via New Clipping dialog.
      cxtMenu.hidePopup();

      // Get the node where the context menu is invoked from.
      var textbox = document.popupNode;
      if (textbox instanceof HTMLInputElement || textbox instanceof HTMLTextAreaElement) {
	var text;
	if (textbox.selectionStart == textbox.selectionEnd) {
	  text = textbox.value;
	}
	else {
	  text = textbox.value.substring(textbox.selectionStart, 
					 textbox.selectionEnd);
	}
	result = this.clippingsSvc.newClippingFromText(text, this.showDialog, window, null, false);
      }
      else {
	var doc = document.popupNode.ownerDocument;

	// ** FIREFOX 2 **  Rich edit field
	if (this.appUtils.isFirefox2() && doc) {
	  cxtMenu.hidePopup();
	  var plainText = doc.defaultView.getSelection();

	  // New (from entire contents)
	  if (plainText == "") {
	    goDoCommand('cmd_selectAll');
	    plainText = doc.defaultView.getSelection();
	    var isTextSelected = true;
	  }
	  result = this.clippingsSvc.newClippingFromText(plainText, this.showDialog, window, null, false);
	  if (isTextSelected) {
	    // TO DO: Deselect text, leaving the cursor where it was.
	  }
	}
	else {
	  this.appUtils.log("In gClippingsExec.newFromTextbox(): textbox = " 
			    + textbox);
	}
      }
    }
    // ** FIREFOX 1.x ONLY **  Rich edit field
    else if (gContextMenu.inFrame) {
      var doc = document.popupNode.ownerDocument;
      cxtMenu.hidePopup();

      if (doc.designMode == "on") {
	var plainText = doc.defaultView.getSelection();

	// New (from entire contents)
	if (plainText == "") {
	  goDoCommand('cmd_selectAll');
	  plainText = doc.defaultView.getSelection();
	  var isTextSelected = true;
	}
      	result = this.clippingsSvc.newClippingFromText(plainText, this.showDialog, window, null, false);
	if (isTextSelected) {
	  // TO DO: Deselect text, leaving the cursor where it was.
	}
      }
    }

    if (result) {
      window.setTimeout("gClippingsExec.saveClippings();", 1);
    }
  },

  newFromSelection: function ()
  {
    if (! this.dataSrcInitialized) {
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }
    var focusedWnd = document.commandDispatcher.focusedWindow;
    var selection = focusedWnd.getSelection();
    if (selection) {
      var result = this.clippingsSvc.newClippingFromText(selection, this.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippingsExec.saveClippings();", 1);
      }
    }
    else {
      var strBundle = document.getElementById("ae-clippings-strings");
      alert(this.strBundle.getString('errorNoSelection'));
    }
  },

  openClippingsManager: function () 
  {
    var wnd = window.open("chrome://clippings/content/clippings.xul",
			  "clippings_wndobj", "chrome,resizable");
    wnd.focus();
  },

  initClippingsPopup: function (aPopup, aMenu) 
  {
    var err = false;
    var dsURL = this.appUtils.getCurrentProfileURL() + this.appUtils.CLIPDAT_FILE_NAME;
    try {
      var ds = this.clippingsSvc.getDataSource(dsURL);
    }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      err = this.strBundle.getString("errorOutOfMemory");
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      err = String.format("%s: %s",
			  this.strBundle.getString("errorAccessDenied"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      err = String.format("%s: %s",
			  this.strBundle.getString("errorFileLocked"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      err = String.format("%s: %s",
			  this.strBundle.getString("errorFileTooBig"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e) {
      // File is corrupt - open Clippings Manager and perform recovery.
      err = 888;
    }

    var errorMenuItem = document.getElementById("ae-clippings-error-menuitem");

    if (err) {
      // Append a new menu item that lets user see more details of error.
      if (errorMenuItem) {
	return;
      }

      var errorCmd = document.createElement("menuitem");
      errorCmd.id = "ae-clippings-error-menuitem";
      errorCmd.setAttribute("label", this.strBundle.getString("errorMenu"));
      errorCmd.style.fontWeight = "bold";

      var func;
      if (err == 888) {
	func = function () { gClippingsExec.openClippingsManager() };
      }
      else {
	func = function () {
	  gClippingsExec.appUtils.alertEx(gClippingsExec.strBundle.getString('appName'), err);
	};
      }
      errorCmd.addEventListener("command", func, false);
      aPopup.appendChild(errorCmd);
      return;
    }
    else {
      // Remove error menu item if error condition no longer exists
      if (errorMenuItem) {
	aPopup.removeChild(errorMenuItem);
      }
    }

    aMenu.database.AddDataSource(ds);
    aMenu.builder.rebuild();

    this.dataSrcInitialized = true;
  },

  insertClippingText: function (aText) 
  {
    this.appUtils.copyTextToClipboard(aText);

    // Paste clipping.  The following function is defined in
    // "chrome://global/content/globalOverlay.js"
    goDoCommand("cmd_paste");
    // SIDE EFFECT: The clipping text will remain on the system clipboard.
  },


  saveClippings: function () 
  {
    var title = this.strBundle.getString('appName');
    try {
      this.clippingsSvc.flushDataSrc();
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_INITIALIZED) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorSaveFailedDSNotInitialized"));
      return;
	   }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorOutOfMemory"));
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      var msg = String.format("%s: %s",
			      this.strBundle.getString("errorAccessDenied"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      var msg = String.format("%s: %s",
			      this.strBundle.getString("errorFileLocked"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      var msg = String.format("%s: %s",
			      this.strBundle.getString("errorFileTooBig"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_READ_ONLY) {
      var msg = String.format("%s: %s",
			      this.strBundle.getString('errorFileReadOnly'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_DISK_FULL) {
      var msg = String.format("%s: %s",
			      this.strBundle.getString('errorDiskFull'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e) {
      this.appUtils.alertEx(title, this.strBundle.getString("alertSaveFailed"));
      return;
    }
  },


  //
  // Browser window and Clippings menu initialization
  //

  isClippingsInitialized: false,

  initClippings: function ()
  {   
    // Workaround to this init function being called multiple times
    if (this.isClippingsInitialized) {
      return;
    }

    this.appUtils = new AppUtils();
    this.strBundle = document.getElementById("ae-clippings-strings");

    try {
      this.clippingsSvc = Components.classes["clippings@mozdev.org/clippings;1"].getService(Components.interfaces.nsIClippingsService);
    }
    catch (e) {
      alert(e);
    }

    // Set up Clippings backup.
    var backupDirURL = this.appUtils.getCurrentProfileURL() + this.appUtils.BACKUP_DIR_NAME;
    this.clippingsSvc.setBackupDir(backupDirURL);
    this.clippingsSvc.setMaxBackupFiles(this.appUtils.getPref('clippings.backup.maxfiles'));

    // First-run initialization - import from Clippings 1.x if necessary.
    if (this.appUtils.getPref("clippings.first_run") == true) {
      this._firstRunInit();
      this.appUtils.setPref("clippings.first_run", false);
    }

    this.appUtils.log("gClippingsExec.initClippings(): Initializing Clippings integration with host app window");

    // Attaching event handlers to browser context menu 
    var hostAppContextMenu = document.getElementById("contentAreaContextMenu");
    hostAppContextMenu.addEventListener("popupshowing", function (e) {gClippingsExec.initContextMenuItem(e)}, false);
    hostAppContextMenu.addEventListener("popuphiding", function (e) {gClippingsExec.hideContextMenuItem(e)}, false);

    // Initializing data source on Clippings context menu
    var menu = document.getElementById("ae-clippings-menu-1");
    var popup = document.getElementById("ae-clippings-popup-1");
    this.initClippingsPopup(popup, menu);

    this.showDialog = !(this.appUtils.getPref("clippings.entries.add_silently"));
    this.isClippingsInitialized = true;

    this.listener = {
      origin:  this.clippingsSvc.ORIGIN_HOSTAPP,

      newFolderCreated: function (aFolderURI)
      {
	this._rebuildMenu();
      },

      newClippingCreated: function (aClippingURI) {},

      importDone: function (aNumItems)
      {
	this._rebuildMenu();
      },

      _rebuildMenu: function ()
      {
	gClippingsExec.appUtils.log("Rebuilding Clippings submenu");
	var menu = document.getElementById("ae-clippings-menu-1");
	menu.builder.rebuild();
	gClippingsExec.appUtils.log("** Rebuild completed successfully **");
      }
    };
    this.clippingsSvc.addListener(this.listener);

    // Initialize "New From Clipboard" command on status bar icon menu.
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var newFromClpbdCmd = document.getElementById("ae_new_clipping_from_clpbd");
    newFromClpbdCmd.setAttribute("label",
				 this.strBundle.getString("newFromClipbd")
				 + ellipsis);
    // Preloading icon images
    this._image1 = document.createElement("image");
    this._image2 = document.createElement("image");
    this._image1.src = "chrome://clippings/skin/images/folder.png";
    this._image2.src = "chrome://clippings/skin/images/clippings-icon.png";
  },

  _image1: null,
  _image2: null,

  _firstRunInit: function ()
  {
    var oldClipdatFile, clipdat2File;
    oldClipdatFile = this.appUtils.getCurrentProfileDir();
    clipdat2File = this.appUtils.getCurrentProfileDir();
    clipdat2File.append(this.appUtils.CLIPDAT_FILE_NAME);
    oldClipdatFile.append(this.appUtils.CLIPDAT_1X_FILE_NAME);
    
    if (!oldClipdatFile.exists() || !oldClipdatFile.isFile()) {
      this.appUtils.log("It appears that Clippings 1.x was never installed previously.  Welcome!");
      return;
    }
    if (clipdat2File.exists() && clipdat2File.isFile()) {
      this.appUtils.log("Clippings data source file already exists.  Skipping first-run initialization.");
      return;
    }

    this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			    document.getElementById("ae-clippings-menu-1"));
    var numImported = -1;
    var oldClipdatURL = this.appUtils.getCurrentProfileURL() + this.appUtils.CLIPDAT_1X_FILE_NAME;
    try {
      numImported = this.clippingsSvc.importFromFile(oldClipdatURL, this.clippingsSvc.FILETYPE_CLIPPINGS_1X);
    }
    catch (e) {
      this.appUtils.alertEx(this.strBundle.getString("appName"),
			    this.strBundle.getString("errorImportFailed"));
      this.appUtils.log(e);
    }

    if (numImported != -1) {
      this.appUtils.log("Your Clippings 1.x data was imported successfully.");

      try {
	this.clippingsSvc.flushDataSrc();
	this.appUtils.log("Flushed imported data to disk.");
      }
      catch (e) {
	alert(this.strBundle.getString("alertSaveFailed") + "\n" + e);
      }
    }
  },


  initContextMenuItem: function (event)
  {
    // WORKAROUND: Although this event handler is attached to the browser's 
    // context menu, it gets called 2 times: for the browser context menu, and 
    // for the Clippings submenu. We only want the following to execute when
    // the browser's context menu is being shown, not the Clippings submenu.
    if (event.target.id != "contentAreaContextMenu") {
      return;
    }

    var popup = document.getElementById("ae-clippings-popup-1");
    var menu = null;

    this.appUtils.log("gClippingsExec.initContextMenuItem(): Initializing Clippings submenu");

    var strBundle = document.getElementById("ae-clippings-strings");
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
    var clippingsMenu2 = document.getElementById("ae-clippings-menu-2");
    var addEntryCmd = document.getElementById("ae_new_clipping_from_textbox");

    // ** FIREFOX 2 **
    // gContextMenu.onTextInput is also true inside a rich edit box!
    if (gContextMenu.onTextInput) {
      var textbox = document.popupNode;

      // Do our own checking to ensure that the node is an HTML text field.
      if (textbox instanceof HTMLInputElement || textbox instanceof HTMLTextAreaElement) {
	if (textbox.selectionStart == textbox.selectionEnd) {
	  // No text selected.  Enable "New..." cmd if text field non-empty
	  addEntryCmd.setAttribute("label",
				   this.strBundle.getString("new") + ellipsis);
	  addEntryCmd.setAttribute("disabled", textbox.value == "");
	}
	else {
	  // "New From Selection..." command.
	  addEntryCmd.setAttribute("label",
				   this.strBundle.getString("newFromSelect") 
				   + ellipsis);
	  addEntryCmd.setAttribute("disabled", "false");
	}
      }
      else {
	var doc = document.popupNode.ownerDocument;

	// ** FIREFOX 2 **  Rich edit box.
	if (this.appUtils.isFirefox2() && doc) {
	  if (gContextMenu.isTextSelected) {
	    clippingsMenu1.hidden = false;
	    clippingsMenu2.hidden = true;
	    menu = clippingsMenu1;
	      
	    addEntryCmd.setAttribute("disabled", "false");
	    addEntryCmd.setAttribute("label", this.strBundle.getString("newFromSelect") + ellipsis);
	  }
	  else {
	    // Check for empty document
	    var range = doc.createRange();
	    range.setStart(doc.body.firstChild, 0);
	    range.setEnd(doc.body.lastChild, 0);
	    addEntryCmd.setAttribute("disabled", range == "");

	    clippingsMenu1.hidden = false;
	    clippingsMenu2.hidden = true;
	    menu = clippingsMenu1;
	    addEntryCmd.setAttribute("label", this.strBundle.getString("new") + ellipsis);
	  }
	}
	else {
	  addEntryCmd.setAttribute("disabled", "true");
	}
      }
      clippingsMenu1.hidden = false;
      clippingsMenu2.hidden = true;
      menu = clippingsMenu1;
    }
    // Selected text in browser content area or in an <IFRAME>
    else if (gContextMenu.isTextSelected) {
      var doc = document.popupNode.ownerDocument;

      // ** FIREFOX 1.x ONLY **
      // Rich edit box
      if (doc.designMode == "on") {
	clippingsMenu1.hidden = false;
	clippingsMenu2.hidden = true;
	menu = clippingsMenu1;

	addEntryCmd.setAttribute("disabled", "false");
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("newFromSelect") 
				 + ellipsis);
      } 
      // Browser content area.
      else {
	clippingsMenu1.hidden = true;
	clippingsMenu2.hidden = false;

	addEntryCmd = document.getElementById("ae_new_clipping_from_selection");
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("newFromSelect") 
				 + ellipsis);
      }
    }
    // ** FIREFOX 1.x ONLY **
    // Rich edit box - if there is any text selected, the preceding
    // "else if ..." statement will be executed.
    else if (gContextMenu.inFrame) {
      var doc = document.popupNode.ownerDocument;
      if (doc.designMode == "on") {
	// Check for empty document
	var range = doc.createRange();
	range.setStart(doc.body.firstChild, 0);
	range.setEnd(doc.body.lastChild, 0);
	addEntryCmd.setAttribute("disabled", range == "");

	clippingsMenu1.hidden = false;
	clippingsMenu2.hidden = true;
	menu = clippingsMenu1;
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("new") + ellipsis);
      }
    }
  },

  hideContextMenuItem: function (event)
  {
    var popup = document.getElementById("ae-clippings-popup-1");
    var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
    var clippingsMenu2 = document.getElementById("ae-clippings-menu-2");

    if (event.target.id == "contentAreaContextMenu") {
      clippingsMenu1.hidden = true;
      clippingsMenu2.hidden = true;
    }
  },

  unload: function ()
  {
    this.clippingsSvc.removeListener(this.listener);
    this.listener = null;
  }
};

// Keep in mind Gecko bug 241518 - calling addEventListener with a closure 
// holding a content node leaks the document (fixed in Firefox 1.6a1)
window.addEventListener("load", function () {gClippingsExec.initClippings()}, false);
window.addEventListener("unload", function () {gClippingsExec.unload()},false);



// Drag 'n drop support
var ae_clippingsStatusBarObserver = {
  onDragOver: function (event, flavour, session)
  {
    event.target.setAttribute("ae-dnd", "true");
  },

  onDragExit: function (event, session)
  {
    event.target.removeAttribute("ae-dnd");
  },

  onDrop: function (event, transferData, session)
  {
    if (! gClippingsExec.dataSrcInitialized) {
      // The RDF data source has to be initialized if it has not already been
      // done so, otherwise RDF node creation will fail, and the new Clippings
      // entry will never be created.
      // This initialization is done in the code for the Clippings popup menu's
      // `onpopupshowing' event handler.
      gClippingsExec.initClippingsPopup(document.getElementById("ae-clippings-popup-1"), document.getElementById("ae-clippings-menu-1"));
    }

    var text = transferData.data;
    var result = gClippingsExec.clippingsSvc.newClippingFromText(text, gClippingsExec.showDialog, window, null, false);
    if (result) {
      window.setTimeout("gClippingsExec.saveClippings();", 100);
    }
  },

  getSupportedFlavours: function ()
  {
    var flavours = new FlavourSet();
    flavours.appendFlavour("text/unicode");
    return flavours;
  }
};


